﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using Phidget22;
using Phidget22.Events;
using System.Runtime.InteropServices;

namespace AdaptiveBackLight
{
    public partial class Form1 : Form
    {
        DigitalOutput redLight = null;
        DigitalOutput greenLight = null;
        DigitalOutput blueLight = null;

        IR ir = null;

        private struct ColorRatios
        {
            public double Red;
            public double Green;
            public double Blue;
        }

        int FOLLOW_SCREEN_MODE = 0;
        int CONSTANT_COLOR_MODE = 1;

        ColorRatios averageColor = new ColorRatios();

        double rMulti;
        double gMulti;
        double bMulti;

        double speed;

        double brightness = 0;

        bool offState = false;
        
        public Form1()
        {
            InitializeComponent();
            this.WindowState = FormWindowState.Minimized;
            Hide();

            redLight = new DigitalOutput();
            greenLight = new DigitalOutput();
            blueLight = new DigitalOutput();
            ir = new IR();

            redLight.Channel = 0;
            greenLight.Channel = 1;
            blueLight.Channel = 2;

            ir.Code += ir_Code;

            redLight.Open();
            greenLight.Open();
            blueLight.Open();
            ir.Open();

            brightness = ((double)brightnessBar.Value / brightnessBar.Maximum);
            rMulti = ((double)redBar.Value / redBar.Maximum);
            gMulti = ((double)greenBar.Value / greenBar.Maximum);
            bMulti = ((double)blueBar.Value / blueBar.Maximum);
            speed = (speedBar.Maximum / (double)speedBar.Value) * 4;

            brightnessLabel.Text = brightness.ToString();
            redLabel.Text = rMulti.ToString();
            greenLabel.Text = gMulti.ToString();
            blueLabel.Text = bMulti.ToString();
            speedLabel.Text = speed.ToString("0.##");

            loadDataFromFile();

            modeBox.SelectedIndex = 0;
        }

        void ir_Code(object sender, IRCodeEventArgs e)
        {
            
            if(String.Equals(e.Code, "20df10ef"))
            {
                if(!e.IsRepeat)
                    offState = !offState;
            }
            else if (String.Equals(e.Code, "00ffb04f"))
            {
                offState = false;
                
            }
            else if (String.Equals(e.Code, "00fff807"))
            {
                offState = true;
            }
            else if (String.Equals(e.Code, "00ff906f"))
            {
                if (brightnessBar.Value <= 90)
                    brightnessBar.Value += 10;
                else
                    brightnessBar.Value = 100;

                brightness = ((double)brightnessBar.Value / brightnessBar.Maximum);
            }
            else if (String.Equals(e.Code, "00ffb847"))
            {
                if (brightnessBar.Value >= 20)
                    brightnessBar.Value -= 10;
                else
                    brightnessBar.Value = 10;

                brightness = ((double)brightnessBar.Value / brightnessBar.Maximum);
            }

            if(offState == false)
            {
                timer1.Enabled = true;
            }
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            getScreenColor();
        }

        private void getScreenColor()
        {
            try
            {
                // Code to capture a bitmap from the screen found at
                // https://stackoverflow.com/questions/362986/capture-the-screen-into-a-bitmap

                //Create a new bitmap.
                Bitmap bmpScreenshot = new Bitmap(Screen.PrimaryScreen.Bounds.Width,
                                       Screen.PrimaryScreen.Bounds.Height,
                                       PixelFormat.Format32bppArgb);

                // Create a graphics object from the bitmap.
                Graphics gfxScreenshot = Graphics.FromImage(bmpScreenshot);

                // Take the screenshot from the upper left corner to the right bottom corner.
                gfxScreenshot.CopyFromScreen(Screen.PrimaryScreen.Bounds.X,
                                                    Screen.PrimaryScreen.Bounds.Y,
                                                    0,
                                                    0,
                                                    Screen.PrimaryScreen.Bounds.Size,
                                                    CopyPixelOperation.SourceCopy);
                //End of Screen Capture Code

                averageColor = CalculateAverageColor(bmpScreenshot);

                bmpScreenshot.Dispose();
                gfxScreenshot.Dispose();

            }
            catch (Exception err)
            {
                averageColor.Red = 0;
                averageColor.Green = 0;
                averageColor.Blue = 0;
            }
        }

        private ColorRatios CalculateAverageColor(Bitmap bm)
        {
            // This code is a modified version of the code found at
            // https://stackoverflow.com/questions/6177499/how-to-determine-the-background-color-of-document-when-there-are-3-options-usin/6185448#6185448
            // which has been specialized for our purposes.

            int width = bm.Width;
            int height = bm.Height;
            int red = 0;
            int green = 0;
            int blue = 0;
            long[] totals = new long[] { 0, 0, 0 };
            int bppModifier = 4; // This will only work for 32-bit bitmaps

            BitmapData srcData = bm.LockBits(new System.Drawing.Rectangle(0, 0, bm.Width, bm.Height), ImageLockMode.ReadOnly, bm.PixelFormat);
            int stride = srcData.Stride;
            IntPtr Scan0 = srcData.Scan0;

            unsafe
            {
                byte* p = (byte*)(void*)Scan0;

                for (int y = 0; y < height; y+=5)
                {
                    for (int x = 0; x < width; x+=5)
                    {
                        int idx = (y * stride) + x * bppModifier;
                        red = p[idx + 2];
                        green = p[idx + 1];
                        blue = p[idx];
                        {
                            totals[2] += red * red;
                            totals[1] += green * green;
                            totals[0] += blue * blue;
                        }
                    }
                }
            }

            int count = (width * height)/25;
            ColorRatios avg = new ColorRatios();
            avg.Red = (totals[2] / count) / 65536.0;
            avg.Green = (totals[1] / count) / 65536.0;
            avg.Blue = (totals[0] / count) / 65536.0;

            //Only start updating the lights once we have colours to work with
            timer2.Enabled = true;

            return avg;
        }

        private void timer2_Tick(object sender, EventArgs e)
        {
            updateLights();
        }

        private void updateLights()
        {
            if (!redLight.Attached || !greenLight.Attached || !blueLight.Attached)
                return;

            if(offState)
            {
                redLight.BeginSetDutyCycle(0, null, null);
                greenLight.BeginSetDutyCycle(0, null, null);
                blueLight.BeginSetDutyCycle(0, null, null);
                timer2.Enabled = false;
                timer1.Enabled = false;
            }
            else if (modeBox.SelectedIndex == FOLLOW_SCREEN_MODE)
            {
                double r = averageColor.Red;
                double g = averageColor.Green;
                double b = averageColor.Blue;

                double screenBrightness = (r + g + b) / 3;

                //allow the lights to go dark if the screen is dark
                if (screenBrightness < 0.1)
                    screenBrightness = 0.1;

                redLight.BeginSetDutyCycle(redLight.DutyCycle + (rMulti * brightness * r / screenBrightness - redLight.DutyCycle) / speed, null, null);
                greenLight.BeginSetDutyCycle(greenLight.DutyCycle + (gMulti * brightness * g / screenBrightness - greenLight.DutyCycle) / speed, null, null);
                blueLight.BeginSetDutyCycle(blueLight.DutyCycle + (bMulti * brightness * b / screenBrightness - blueLight.DutyCycle) / speed, null, null);
            }
            else if (modeBox.SelectedIndex == CONSTANT_COLOR_MODE)
            { 
                redLight.BeginSetDutyCycle(rMulti * brightness, null, null);
                greenLight.BeginSetDutyCycle(gMulti * brightness, null, null);
                blueLight.BeginSetDutyCycle(bMulti * brightness, null, null);
            }
        }

        private void redBar_Scroll(object sender, EventArgs e)
        {
            rMulti = ((double)redBar.Value / redBar.Maximum);
            redLabel.Text = rMulti.ToString();
        }

        private void greenBar_Scroll(object sender, EventArgs e)
        {
            gMulti = ((double)greenBar.Value / greenBar.Maximum);
            greenLabel.Text = gMulti.ToString();
        }

        private void blueBar_Scroll(object sender, EventArgs e)
        {
            bMulti = ((double)blueBar.Value / blueBar.Maximum);
            blueLabel.Text = bMulti.ToString();
        }

        private void brightnessBar_Scroll(object sender, EventArgs e)
        {
            brightness = ((double)brightnessBar.Value / brightnessBar.Maximum);
            brightnessLabel.Text = brightness.ToString();
            if (brightness == 0)
                offState = true;
            else {
                offState = false;
                timer1.Enabled = true;
            }
        }

        private void speedBar_Scroll(object sender, EventArgs e)
        {
            speed = (speedBar.Maximum / (double)speedBar.Value) * 4;
            speedLabel.Text = speed.ToString("0.##");

        }

        private void modeBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (modeBox.SelectedIndex == FOLLOW_SCREEN_MODE && !offState)
                timer1.Enabled = true;
            else
                timer1.Enabled = false; //stop capturing the screen if we don't need to
        }

        private void notifyIcon1_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            Show();
            Activate();
            WindowState = FormWindowState.Normal;
            ShowInTaskbar = true;
        }

        private void Form1_Resize(object sender, EventArgs e)
        {
            if (FormWindowState.Minimized == WindowState)
            {
                Hide();
                ShowInTaskbar = false;
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            redLight.Close();
            greenLight.Close();
            blueLight.Close();

            notifyIcon1.Dispose();
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            String[] settings = new String[6];

            settings[0] = rMulti.ToString();
            settings[1] = gMulti.ToString();
            settings[2] = bMulti.ToString();

            settings[3] = brightness.ToString();
            settings[4] = speed.ToString("0.##");
            settings[5] = modeBox.SelectedIndex.ToString();
            System.IO.File.WriteAllLines("adaptiveLightSettings.cfg", settings);
        }

        private void loadDataFromFile()
        {
            try {
                String[] settings = System.IO.File.ReadAllLines("adaptiveLightSettings.cfg");

                rMulti = Convert.ToDouble(settings[0]);
                gMulti = Convert.ToDouble(settings[1]);
                bMulti = Convert.ToDouble(settings[2]);

                brightness = Convert.ToDouble(settings[3]);
                speed = Convert.ToDouble(settings[4]);
                modeBox.SelectedIndex = Convert.ToInt32(settings[5]);

                brightnessBar.Value = (int)(brightnessBar.Maximum * brightness);
                speedBar.Value = (int)(Math.Round((speedBar.Maximum / speed) * 4));
                redBar.Value = (int)(redBar.Maximum * rMulti);
                greenBar.Value = (int)(greenBar.Maximum * gMulti);
                blueBar.Value = (int)(blueBar.Maximum * bMulti);

                brightnessLabel.Text = brightness.ToString();
                redLabel.Text = rMulti.ToString();
                greenLabel.Text = gMulti.ToString();
                blueLabel.Text = bMulti.ToString();
                speedLabel.Text = speed.ToString("0.##");
            }
            catch (Exception e)
            {
                MessageBox.Show("Failed to load settings:\n" + e.Message);
            }
        }

        private void loadButton_Click(object sender, EventArgs e)
        {
            loadDataFromFile();
        }
    }

}
